@@ -13,9 +13,6 @@
 ##############################################################################
 """Database objects
 """
-
-import cPickle
-import cStringIO
 import sys
 import threading
 import logging
@@ -26,17 +23,19 @@ import warnings
 from ZODB.broken import find_global
 from ZODB.utils import z64
 from ZODB.Connection import Connection
+from ZODB._compat import Pickler, _protocol, BytesIO
 import ZODB.serialize
 
 import transaction.weakset
 
-from zope.interface import implements
+from zope.interface import implementer
 from ZODB.interfaces import IDatabase
 from ZODB.interfaces import IMVCCStorage
 
 import transaction
 
 from persistent.TimeStamp import TimeStamp
+import six
 
 
 logger = logging.getLogger('ZODB.DB')
@@ -146,7 +145,7 @@ class ConnectionPool(AbstractConnectionPool):
         n = len(self.all)
         limit = self.size
         if n > limit:
-            reporter = logger.warn
+            reporter = logger.warning
             if n > 2 * limit:
                 reporter = logger.critical
             reporter("DB.open() has %s open connections with a pool_size "
@@ -271,7 +270,7 @@ class KeyedConnectionPool(AbstractConnectionPool):
             return pool.pop()
 
     def map(self, f):
-        for pool in self.pools.itervalues():
+        for pool in six.itervalues(self.pools):
             pool.map(f)
 
     def availableGC(self):
@@ -283,14 +282,14 @@ class KeyedConnectionPool(AbstractConnectionPool):
     @property
     def test_all(self):
         result = set()
-        for pool in self.pools.itervalues():
+        for pool in six.itervalues(self.pools):
             result.update(pool.all)
         return frozenset(result)
 
     @property
     def test_available(self):
         result = []
-        for pool in self.pools.itervalues():
+        for pool in six.itervalues(self.pools):
             result.extend(pool.available)
         return tuple(result)
 
@@ -310,15 +309,16 @@ def getTID(at, before):
             at = toTimeStamp(at)
         else:
             at = TimeStamp(at)
-        before = repr(at.laterThan(at))
+        before = at.laterThan(at).raw()
     elif before is not None:
         if isinstance(before, datetime.datetime):
-            before = repr(toTimeStamp(before))
+            before = toTimeStamp(before).raw()
         else:
-            before = repr(TimeStamp(before))
+            before = TimeStamp(before).raw()
     return before
 
 
+@implementer(IDatabase)
 class DB(object):
     """The Object Database
     -------------------
@@ -359,7 +359,6 @@ class DB(object):
         cacheDetailSize, getCacheSize, getHistoricalCacheSize, setCacheSize,
         setHistoricalCacheSize
     """
-    implements(IDatabase)
 
     klass = Connection  # Class to use for connections
     _activity_monitor = next = previous = None
@@ -398,7 +397,7 @@ class DB(object):
           - `xrefs` - Boolian flag indicating whether implicit cross-database
             references are allowed
         """
-        if isinstance(storage, basestring):
+        if isinstance(storage, six.string_types):
             from ZODB import FileStorage
             storage = ZODB.FileStorage.FileStorage(storage, **storage_args)
         elif storage is None:
@@ -448,8 +447,8 @@ class DB(object):
                 root = PersistentMapping()
                 # Manually create a pickle for the root to put in the storage.
                 # The pickle must be in the special ZODB format.
-                file = cStringIO.StringIO()
-                p = cPickle.Pickler(file, 1)
+                file = BytesIO()
+                p = Pickler(file, _protocol)
                 p.dump((root.__class__, None))
                 p.dump(root.__getstate__())
                 t = transaction.Transaction()
@@ -491,10 +490,6 @@ class DB(object):
             assert connection._db is self
             connection.opened = None
 
-            am = self._activity_monitor
-            if am is not None:
-                am.closedConnection(connection)
-
             if connection.before:
                 self.historical_pool.repush(connection, connection.before)
             else:
@@ -530,9 +525,7 @@ class DB(object):
                     detail[c] = 1
 
         self._connectionMap(f)
-        detail = detail.items()
-        detail.sort()
-        return detail
+        return sorted(detail.items())
 
     def cacheExtremeDetail(self):
         detail = []
@@ -544,9 +537,9 @@ class DB(object):
                 id = ''
                 if hasattr(ob, '__dict__'):
                     d = ob.__dict__
-                    if d.has_key('id'):
+                    if 'id' in d:
                         id = d['id']
-                    elif d.has_key('__name__'):
+                    elif '__name__' in d:
                         id = d['__name__']
 
                 module = getattr(ob.__class__, '__module__', '')
@@ -606,8 +599,9 @@ class DB(object):
                       'ngsize': con._cache.cache_non_ghost_count,
                       'size': len(con._cache)})
         self._connectionMap(f)
-        m.sort()
-        return m
+        # Py3: Simulate Python 2 m.sort() functionality.
+        return sorted(
+            m, key=lambda x: (x['connection'], x['ngsize'], x['size']))
 
     def close(self):
         """Close the database and its underlying storage.
@@ -717,7 +711,7 @@ class DB(object):
             raise ValueError(
                 'cannot open an historical connection in the future.')
 
-        if isinstance(transaction_manager, basestring):
+        if isinstance(transaction_manager, six.string_types):
             if transaction_manager:
                 raise TypeError("Versions aren't supported.")
             warnings.warn(
@@ -927,7 +921,7 @@ class DB(object):
             raise NotImplementedError
         if txn is None:
             txn = transaction.get()
-        if isinstance(ids, basestring):
+        if isinstance(ids, six.string_types):
             ids = [ids]
         txn.join(TransactionalUndo(self, ids))
 
@@ -956,6 +950,15 @@ class DB(object):
     def new_oid(self):
         return self.storage.new_oid()
 
+    def open_then_close_db_when_connection_closes(self):
+        """Create and return a connection.
+
+        When the connection closes, the database will close too.
+        """
+        conn = self.open()
+        conn.onCloseCallback(self.close)
+        return conn
+
 
 class ContextManager:
     """PEP 343 context manager
@@ -1016,7 +1019,4 @@ class TransactionalUndo(object):
         return "%s:%s" % (self._storage.sortKey(), id(self))
 
 def connection(*args, **kw):
-    db = DB(*args, **kw)
-    conn = db.open()
-    conn.onCloseCallback(db.close)
-    return conn
+    return DB(*args, **kw).open_then_close_db_when_connection_closes()
