@@ -14,7 +14,6 @@
 """Blobs
 """
 
-import base64
 import binascii
 import logging
 import os
@@ -23,22 +22,25 @@ import shutil
 import stat
 import sys
 import tempfile
-import threading
-import time
 import weakref
 
 import zope.interface
+import persistent
 
 import ZODB.interfaces
 from ZODB.interfaces import BlobError
 from ZODB import utils
 from ZODB.POSException import POSKeyError
-import transaction
-import transaction.interfaces
-import persistent
+from ZODB._compat import BytesIO
+from ZODB._compat import Unpickler
+from ZODB._compat import decodebytes
+from ZODB._compat import ascii_bytes
+from ZODB._compat import INT_TYPES
+
+
+if sys.version_info[0] >= 3:
+    from io import FileIO as file
 
-from zope.proxy import getProxiedObject, non_overridable
-from zope.proxy.decorator import SpecificationDecoratorBase
 
 logger = logging.getLogger('ZODB.blob')
 
@@ -48,7 +50,7 @@ SAVEPOINT_SUFFIX = ".spb"
 LAYOUT_MARKER = '.layout'
 LAYOUTS = {}
 
-valid_modes = 'r', 'w', 'r+', 'a'
+valid_modes = 'r', 'w', 'r+', 'a', 'c'
 
 # Threading issues:
 # We want to support closing blob files when they are destroyed.
@@ -56,22 +58,25 @@ valid_modes = 'r', 'w', 'r+', 'a'
 # via GC in any thread.
 
 
+@zope.interface.implementer(ZODB.interfaces.IBlob)
 class Blob(persistent.Persistent):
     """A BLOB supports efficient handling of large data within ZODB."""
 
-    zope.interface.implements(ZODB.interfaces.IBlob)
 
     _p_blob_uncommitted = None  # Filename of the uncommitted (dirty) data
     _p_blob_committed = None    # Filename of the committed data
 
     readers = writers = None
 
-    def __init__(self):
+    def __init__(self, data=None):
         # Raise exception if Blobs are getting subclassed
         # refer to ZODB-Bug No.127182 by Jim Fulton on 2007-07-20
         if (self.__class__ is not Blob):
             raise TypeError('Blobs do not support subclassing.')
         self.__setstate__()
+        if data is not None:
+            with self.open('w') as f:
+                f.write(data)
 
     def __setstate__(self, state=None):
         # we use lists here because it will allow us to add and remove
@@ -122,6 +127,17 @@ class Blob(persistent.Persistent):
         if mode not in valid_modes:
             raise ValueError("invalid mode", mode)
 
+        if mode == 'c':
+            if (self._p_blob_uncommitted
+                or
+                not self._p_blob_committed
+                or
+                self._p_blob_committed.endswith(SAVEPOINT_SUFFIX)
+                ):
+                raise BlobError('Uncommitted changes')
+            return self._p_jar._storage.openCommittedBlobFile(
+                self._p_oid, self._p_serial)
+
         if self.writers:
             raise BlobError("Already opened for writing.")
 
@@ -129,10 +145,20 @@ class Blob(persistent.Persistent):
             self.readers = []
 
         if mode == 'r':
-            if self._current_filename() is None:
-                self._create_uncommitted_file()
+            result = None
+            to_open = self._p_blob_uncommitted
+            if not to_open:
+                to_open = self._p_blob_committed
+                if to_open:
+                    result = self._p_jar._storage.openCommittedBlobFile(
+                        self._p_oid, self._p_serial, self)
+                else:
+                    self._create_uncommitted_file()
+                    to_open = self._p_blob_uncommitted
+                    assert to_open
 
-            result = BlobFile(self._current_filename(), mode, self)
+            if result is None:
+                result = BlobFile(to_open, mode, self)
 
             def destroyed(ref, readers=self.readers):
                 try:
@@ -149,14 +175,16 @@ class Blob(persistent.Persistent):
                 if self._p_blob_uncommitted is None:
                     self._create_uncommitted_file()
                 result = BlobFile(self._p_blob_uncommitted, mode, self)
-            else:
+            else: # 'r+' and 'a'
                 if self._p_blob_uncommitted is None:
                     # Create a new working copy
                     self._create_uncommitted_file()
                     result = BlobFile(self._p_blob_uncommitted, mode, self)
-                    utils.cp(file(self._p_blob_committed), result)
-                    if mode == 'r+':
-                        result.seek(0)
+                    if self._p_blob_committed:
+                        with open(self._p_blob_committed, 'rb') as fp:
+                            utils.cp(fp, result)
+                        if mode == 'r+':
+                            result.seek(0)
                 else:
                     # Re-use existing working copy
                     result = BlobFile(self._p_blob_uncommitted, mode, self)
@@ -181,7 +209,15 @@ class Blob(persistent.Persistent):
             self._p_blob_committed.endswith(SAVEPOINT_SUFFIX)
             ):
             raise BlobError('Uncommitted changes')
-        return self._p_blob_committed
+
+        result = self._p_blob_committed
+
+        # We do this to make sure we have the file and to let the
+        # storage know we're accessing the file.
+        n = self._p_jar._storage.loadBlob(self._p_oid, self._p_serial)
+        assert result == n, (result, n)
+
+        return result
 
     def consumeFile(self, filename):
         """Will replace the current data of the blob with the file given under
@@ -224,7 +260,7 @@ class Blob(persistent.Persistent):
             raise
         else:
             if previous_uncommitted:
-                # The relinking worked so we can remove the data that we had 
+                # The relinking worked so we can remove the data that we had
                 # set aside.
                 os.remove(target_aside)
 
@@ -234,11 +270,6 @@ class Blob(persistent.Persistent):
 
     # utility methods
 
-    def _current_filename(self):
-        # NOTE: _p_blob_committed and _p_blob_uncommitted appear by virtue of
-        # Connection._setstate
-        return self._p_blob_uncommitted or self._p_blob_committed
-
     def _create_uncommitted_file(self):
         assert self._p_blob_uncommitted is None, (
             "Uncommitted file already exists.")
@@ -246,7 +277,7 @@ class Blob(persistent.Persistent):
             tempdir = self._p_jar.db()._storage.temporaryDirectory()
         else:
             tempdir = tempfile.gettempdir()
-        filename = utils.mktemp(dir=tempdir)
+        filename = utils.mktemp(dir=tempdir, prefix="BUC")
         self._p_blob_uncommitted = filename
 
         def cleanup(ref):
@@ -284,7 +315,7 @@ class BlobFile(file):
 
     def close(self):
         self.blob.closed(self)
-        file.close(self)
+        super(BlobFile, self).close()
 
 _pid = str(os.getpid())
 
@@ -301,7 +332,7 @@ class FilesystemHelper:
     # want to perform blob storage differently.
 
     def __init__(self, base_dir, layout_name='automatic'):
-        self.base_dir = os.path.normpath(base_dir) + os.path.sep
+        self.base_dir = os.path.abspath(base_dir) + os.path.sep
         self.temp_dir = os.path.join(base_dir, 'tmp')
 
         if layout_name == 'automatic':
@@ -310,26 +341,26 @@ class FilesystemHelper:
             log('The `lawn` blob directory layout is deprecated due to '
                 'scalability issues on some file systems, please consider '
                 'migrating to the `bushy` layout.', level=logging.WARN)
-        self.layout_name = layout_name 
+        self.layout_name = layout_name
         self.layout = LAYOUTS[layout_name]
 
     def create(self):
         if not os.path.exists(self.base_dir):
-            os.makedirs(self.base_dir, 0700)
+            os.makedirs(self.base_dir, 0o700)
             log("Blob directory '%s' does not exist. "
                 "Created new directory." % self.base_dir)
         if not os.path.exists(self.temp_dir):
-            os.makedirs(self.temp_dir, 0700)
+            os.makedirs(self.temp_dir, 0o700)
             log("Blob temporary directory '%s' does not exist. "
                 "Created new directory." % self.temp_dir)
 
-        if not os.path.exists(os.path.join(self.base_dir, LAYOUT_MARKER)):
-            layout_marker = open(
-                os.path.join(self.base_dir, LAYOUT_MARKER), 'wb')
-            layout_marker.write(self.layout_name)
+        layout_marker_path = os.path.join(self.base_dir, LAYOUT_MARKER)
+        if not os.path.exists(layout_marker_path):
+            with open(layout_marker_path, 'w') as layout_marker:
+                layout_marker.write(self.layout_name)
         else:
-            layout = open(os.path.join(self.base_dir, LAYOUT_MARKER), 'rb'
-                          ).read().strip()
+            with open(layout_marker_path, 'r') as layout_marker:
+                layout = layout_marker.read().strip()
             if layout != self.layout_name:
                 raise ValueError(
                     "Directory layout `%s` selected for blob directory %s, but "
@@ -338,7 +369,7 @@ class FilesystemHelper:
 
     def isSecure(self, path):
         """Ensure that (POSIX) path mode bits are 0700."""
-        return (os.stat(path).st_mode & 077) == 0
+        return (os.stat(path).st_mode & 0o77) == 0
 
     def checkSecure(self):
         if not self.isSecure(self.base_dir):
@@ -363,7 +394,7 @@ class FilesystemHelper:
 
         if create and not os.path.exists(path):
             try:
-                os.makedirs(path, 0700)
+                os.makedirs(path, 0o700)
             except OSError:
                 # We might have lost a race.  If so, the directory
                 # must exist now
@@ -380,18 +411,26 @@ class FilesystemHelper:
         path = path[len(self.base_dir):]
         return self.layout.path_to_oid(path)
 
+    def createPathForOID(self, oid):
+        """Given an OID, creates a directory on the filesystem where
+        the blob data relating to that OID is stored, if it doesn't exist.
+        """
+        return self.getPathForOID(oid, create=True)
+
     def getBlobFilename(self, oid, tid):
         """Given an oid and a tid, return the full filename of the
         'committed' blob file related to that oid and tid.
 
         """
-        oid_path = self.getPathForOID(oid)
         # TIDs are numbers and sometimes passed around as integers. For our
         # computations we rely on the 64-bit packed string representation
+        if isinstance(oid, int):
+            oid = utils.p64(oid)
         if isinstance(tid, int):
             tid = utils.p64(tid)
-        filename = "%s%s" % (utils.tid_repr(tid), BLOB_SUFFIX)
-        return os.path.join(oid_path, filename)
+        return os.path.join(self.base_dir,
+                            self.layout.getBlobFilePath(oid, tid),
+                            )
 
     def blob_mkstemp(self, oid, tid):
         """Given an oid and a tid, return a temporary file descriptor
@@ -405,7 +444,8 @@ class FilesystemHelper:
 
         """
         oidpath = self.getPathForOID(oid)
-        fd, name = tempfile.mkstemp(suffix='.tmp', prefix=utils.tid_repr(tid),
+        fd, name = tempfile.mkstemp(suffix='.tmp',
+                                    prefix=utils.tid_repr(tid),
                                     dir=oidpath)
         return fd, name
 
@@ -423,7 +463,7 @@ class FilesystemHelper:
 
         serial = filename[:-len(BLOB_SUFFIX)]
         serial = utils.repr_to_oid(serial)
-        return oid, serial 
+        return oid, serial
 
     def getOIDsForSerial(self, search_serial):
         """Return all oids related to a particular tid that exist in
@@ -455,6 +495,15 @@ class FilesystemHelper:
             yield oid, path
 
 
+class NoBlobsFileSystemHelper:
+
+    @property
+    def temp_dir(self):
+        raise TypeError("Blobs are not supported")
+
+    getPathForOID = getBlobFilename = temp_dir
+
+
 class BlobStorageError(Exception):
     """The blob storage encountered an invalid state."""
 
@@ -462,23 +511,30 @@ def auto_layout_select(path):
     # A heuristic to look at a path and determine which directory layout to
     # use.
     layout_marker = os.path.join(path, LAYOUT_MARKER)
-    if not os.path.exists(path):
+    if os.path.exists(layout_marker):
+        with open(layout_marker, 'r') as fp:
+            layout = fp.read().strip()
+        log('Blob directory `%s` has layout marker set. '
+            'Selected `%s` layout. ' % (path, layout), level=logging.DEBUG)
+    elif not os.path.exists(path):
         log('Blob directory %s does not exist. '
             'Selected `bushy` layout. ' % path)
         layout = 'bushy'
-    elif len(os.listdir(path)) == 0:
-        log('Blob directory `%s` is unused and has no layout marker set. '
-            'Selected `bushy` layout. ' % path)
-        layout = 'bushy'
-    elif LAYOUT_MARKER not in os.listdir(path):
-        log('Blob directory `%s` is used but has no layout marker set. '
-            'Selected `lawn` layout. ' % path)
-        layout = 'lawn'
     else:
-        layout = open(layout_marker, 'rb').read()
-        layout = layout.strip()
-        log('Blob directory `%s` has layout marker set. '
-            'Selected `%s` layout. ' % (path, layout))
+        # look for a non-hidden file in the directory
+        has_files = False
+        for name in os.listdir(path):
+            if not name.startswith('.'):
+                has_files = True
+                break
+        if not has_files:
+            log('Blob directory `%s` is unused and has no layout marker set. '
+                'Selected `bushy` layout. ' % path)
+            layout = 'bushy'
+        else:
+            log('Blob directory `%s` is used but has no layout marker set. '
+                'Selected `lawn` layout. ' % path)
+            layout = 'lawn'
     return layout
 
 
@@ -497,23 +553,37 @@ class BushyLayout(object):
         directories = []
         # Create the bushy directory structure with the least significant byte
         # first
-        for byte in str(oid):
-            directories.append('0x%s' % binascii.hexlify(byte))
+        for byte in ascii_bytes(oid):
+            if isinstance(byte,INT_TYPES): # Py3k iterates byte strings as ints
+                hex_segment_bytes = b'0x' + binascii.hexlify(bytes([byte]))
+                hex_segment_string = hex_segment_bytes.decode('ascii')
+            else:
+                hex_segment_string = '0x%s' % binascii.hexlify(byte)
+            directories.append(hex_segment_string)
+
         return os.path.sep.join(directories)
 
     def path_to_oid(self, path):
         if self.blob_path_pattern.match(path) is None:
             raise ValueError("Not a valid OID path: `%s`" % path)
-        path = path.split(os.path.sep)
+        path = [ascii_bytes(x) for x in path.split(os.path.sep)]
         # Each path segment stores a byte in hex representation. Turn it into
         # an int and then get the character for our byte string.
-        oid = ''.join(binascii.unhexlify(byte[2:]) for byte in path)
+        oid = b''.join(binascii.unhexlify(byte[2:]) for byte in path)
         return oid
 
+    def getBlobFilePath(self, oid, tid):
+        """Given an oid and a tid, return the full filename of the
+        'committed' blob file related to that oid and tid.
+
+        """
+        oid_path = self.oid_to_path(oid)
+        filename = "%s%s" % (utils.tid_repr(tid), BLOB_SUFFIX)
+        return os.path.join(oid_path, filename)
+
 LAYOUTS['bushy'] = BushyLayout()
 
-
-class LawnLayout(object):
+class LawnLayout(BushyLayout):
     """A shallow directory layout for blob directories.
 
     Creates a single level of directories (one for each oid).
@@ -530,57 +600,83 @@ class LawnLayout(object):
                 # OID z64.
                 raise TypeError()
             return utils.repr_to_oid(path)
-        except TypeError:
+        except (TypeError, binascii.Error):
             raise ValueError('Not a valid OID path: `%s`' % path)
 
 LAYOUTS['lawn'] = LawnLayout()
 
-class BlobStorage(SpecificationDecoratorBase):
-    """A storage to support blobs."""
+class BlobStorageMixin(object):
+    """A mix-in to help storages support blobs."""
 
-    zope.interface.implements(ZODB.interfaces.IBlobStorage)
-
-    # Proxies can't have a __dict__ so specifying __slots__ here allows
-    # us to have instance attributes explicitly on the proxy.
-    __slots__ = ('fshelper', 'dirty_oids', '_BlobStorage__supportsUndo',
-                 '_blobs_pack_is_in_progress', )
-
-    def __new__(self, base_directory, storage, layout='automatic'):
-        return SpecificationDecoratorBase.__new__(self, storage)
-
-    def __init__(self, base_directory, storage, layout='automatic'):
+    def _blob_init(self, blob_dir, layout='automatic'):
         # XXX Log warning if storage is ClientStorage
-        SpecificationDecoratorBase.__init__(self, storage)
-        self.fshelper = FilesystemHelper(base_directory, layout)
+        self.fshelper = FilesystemHelper(blob_dir, layout)
         self.fshelper.create()
         self.fshelper.checkSecure()
         self.dirty_oids = []
+
+    def _blob_init_no_blobs(self):
+        self.fshelper = NoBlobsFileSystemHelper()
+        self.dirty_oids = []
+
+    def _blob_tpc_abort(self):
+        """Blob cleanup to be called from subclass tpc_abort
+        """
+        while self.dirty_oids:
+            oid, serial = self.dirty_oids.pop()
+            clean = self.fshelper.getBlobFilename(oid, serial)
+            if os.path.exists(clean):
+                remove_committed(clean)
+
+    def _blob_tpc_finish(self):
+        """Blob cleanup to be called from subclass tpc_finish
+        """
+        self.dirty_oids = []
+
+    def registerDB(self, db):
+        self.__untransform_record_data = db.untransform_record_data
         try:
-            supportsUndo = storage.supportsUndo
+            m = super(BlobStorageMixin, self).registerDB
         except AttributeError:
-            supportsUndo = False
+            pass
         else:
-            supportsUndo = supportsUndo()
-        self.__supportsUndo = supportsUndo
-        self._blobs_pack_is_in_progress = False
+            m(db)
 
-    @non_overridable
-    def temporaryDirectory(self):
-        return self.fshelper.temp_dir
+    def __untransform_record_data(self, record):
+        return record
 
-    @non_overridable
-    def __repr__(self):
-        normal_storage = getProxiedObject(self)
-        return '<BlobStorage proxy for %r at %s>' % (normal_storage,
-                                                     hex(id(self)))
-    @non_overridable
-    def storeBlob(self, oid, oldserial, data, blobfilename, version,
-                  transaction):
-        """Stores data that has a BLOB attached."""
-        serial = self.store(oid, oldserial, data, version, transaction)
-        assert isinstance(serial, str) # XXX in theory serials could be 
-                                       # something else
+    def is_blob_record(self, record):
+        if record:
+            return is_blob_record(self.__untransform_record_data(record))
 
+    def copyTransactionsFrom(self, other):
+        copyTransactionsFromTo(other, self)
+
+    def loadBlob(self, oid, serial):
+        """Return the filename where the blob file can be found.
+        """
+        filename = self.fshelper.getBlobFilename(oid, serial)
+        if not os.path.exists(filename):
+            raise POSKeyError("No blob file at %s" % filename, oid, serial)
+        return filename
+
+    def openCommittedBlobFile(self, oid, serial, blob=None):
+        blob_filename = self.loadBlob(oid, serial)
+        if blob is None:
+            return open(blob_filename, 'rb')
+        else:
+            return BlobFile(blob_filename, 'r', blob)
+
+    def restoreBlob(self, oid, serial, data, blobfilename, prev_txn,
+                    transaction):
+        """Write blob data already committed in a separate database
+        """
+        self.restore(oid, serial, data, '', prev_txn, transaction)
+        self._blob_storeblob(oid, serial, blobfilename)
+
+        return self._tid
+
+    def _blob_storeblob(self, oid, serial, blobfilename):
         self._lock_acquire()
         try:
             self.fshelper.getPathForOID(oid, create=True)
@@ -592,38 +688,73 @@ class BlobStorage(SpecificationDecoratorBase):
             self.dirty_oids.append((oid, serial))
         finally:
             self._lock_release()
+
+    def storeBlob(self, oid, oldserial, data, blobfilename, version,
+                  transaction):
+        """Stores data that has a BLOB attached."""
+        assert not version, "Versions aren't supported."
+        serial = self.store(oid, oldserial, data, '', transaction)
+        self._blob_storeblob(oid, serial, blobfilename)
+
         return self._tid
 
-    @non_overridable
+    def temporaryDirectory(self):
+        return self.fshelper.temp_dir
+
+
+@zope.interface.implementer(ZODB.interfaces.IBlobStorage)
+class BlobStorage(BlobStorageMixin):
+    """A wrapper/proxy storage to support blobs.
+    """
+
+
+    def __init__(self, base_directory, storage, layout='automatic'):
+        assert not ZODB.interfaces.IBlobStorage.providedBy(storage)
+        self.__storage = storage
+
+        self._blob_init(base_directory, layout)
+        try:
+            supportsUndo = storage.supportsUndo
+        except AttributeError:
+            supportsUndo = False
+        else:
+            supportsUndo = supportsUndo()
+        self.__supportsUndo = supportsUndo
+        self._blobs_pack_is_in_progress = False
+
+        if ZODB.interfaces.IStorageRestoreable.providedBy(storage):
+            iblob = ZODB.interfaces.IBlobStorageRestoreable
+        else:
+            iblob = ZODB.interfaces.IBlobStorage
+
+        zope.interface.directlyProvides(
+            self, iblob, zope.interface.providedBy(storage))
+
+    def __getattr__(self, name):
+        return getattr(self.__storage, name)
+
+    def __len__(self):
+        return len(self.__storage)
+
+    def __repr__(self):
+        normal_storage = self.__storage
+        return '<BlobStorage proxy for %r at %s>' % (normal_storage,
+                                                     hex(id(self)))
+
     def tpc_finish(self, *arg, **kw):
         # We need to override the base storage's tpc_finish instead of
-        # providing a _finish method because methods found on the proxied 
+        # providing a _finish method because methods found on the proxied
         # object aren't rebound to the proxy
-        getProxiedObject(self).tpc_finish(*arg, **kw)
-        self.dirty_oids = []
+        self.__storage.tpc_finish(*arg, **kw)
+        self._blob_tpc_finish()
 
-    @non_overridable
     def tpc_abort(self, *arg, **kw):
         # We need to override the base storage's abort instead of
         # providing an _abort method because methods found on the proxied object
         # aren't rebound to the proxy
-        getProxiedObject(self).tpc_abort(*arg, **kw)
-        while self.dirty_oids:
-            oid, serial = self.dirty_oids.pop()
-            clean = self.fshelper.getBlobFilename(oid, serial)
-            if os.path.exists(clean):
-                remove_committed(clean) 
+        self.__storage.tpc_abort(*arg, **kw)
+        self._blob_tpc_abort()
 
-    @non_overridable
-    def loadBlob(self, oid, serial):
-        """Return the filename where the blob file can be found.
-        """
-        filename = self.fshelper.getBlobFilename(oid, serial)
-        if not os.path.exists(filename):
-            raise POSKeyError("No blob file", oid, serial)
-        return filename
-
-    @non_overridable
     def _packUndoing(self, packtime, referencesf):
         # Walk over all existing revisions of all blob files and check
         # if they are still needed by attempting to load the revision
@@ -631,7 +762,6 @@ class BlobStorage(SpecificationDecoratorBase):
         # possible way to do this, but it's safe.
         for oid, oid_path in self.fshelper.listOIDs():
             files = os.listdir(oid_path)
-
             for filename in files:
                 filepath = os.path.join(oid_path, filename)
                 whatever, serial = self.fshelper.splitBlobFilename(filepath)
@@ -643,7 +773,6 @@ class BlobStorage(SpecificationDecoratorBase):
             if not os.listdir(oid_path):
                 shutil.rmtree(oid_path)
 
-    @non_overridable
     def _packNonUndoing(self, packtime, referencesf):
         for oid, oid_path in self.fshelper.listOIDs():
             exists = True
@@ -657,8 +786,8 @@ class BlobStorage(SpecificationDecoratorBase):
                 files.sort()
                 latest = files[-1] # depends on ever-increasing tids
                 files.remove(latest)
-                for file in files:
-                    remove_committed(os.path.join(oid_path, file))
+                for f in files:
+                    remove_committed(os.path.join(oid_path, f))
             else:
                 remove_committed_dir(oid_path)
                 continue
@@ -666,7 +795,6 @@ class BlobStorage(SpecificationDecoratorBase):
             if not os.listdir(oid_path):
                 shutil.rmtree(oid_path)
 
-    @non_overridable
     def pack(self, packtime, referencesf):
         """Remove all unused OID/TID combinations."""
         self._lock_acquire()
@@ -680,7 +808,7 @@ class BlobStorage(SpecificationDecoratorBase):
         try:
             # Pack the underlying storage, which will allow us to determine
             # which serials are current.
-            unproxied = getProxiedObject(self)
+            unproxied = self.__storage
             result = unproxied.pack(packtime, referencesf)
 
             # Perform a pack on the blob data.
@@ -695,16 +823,8 @@ class BlobStorage(SpecificationDecoratorBase):
 
         return result
 
-    @non_overridable
-    def getSize(self):
-        """Return the size of the database in bytes."""
-        # XXX The old way of computing is way to resource hungry. We need to
-        # do some kind of estimation instead.
-        return getProxiedObject(self).getSize()
-
-    @non_overridable
     def undo(self, serial_id, transaction):
-        undo_serial, keys = getProxiedObject(self).undo(serial_id, transaction)
+        undo_serial, keys = self.__storage.undo(serial_id, transaction)
         # serial_id is the transaction id of the txn that we wish to undo.
         # "undo_serial" is the transaction id of txn in which the undo is
         # performed.  "keys" is the list of oids that are involved in the
@@ -712,7 +832,7 @@ class BlobStorage(SpecificationDecoratorBase):
 
         # The serial_id is assumed to be given to us base-64 encoded
         # (belying the web UI legacy of the ZODB code :-()
-        serial_id = base64.decodestring(serial_id+'\n')
+        serial_id = decodebytes(serial_id + b'\n')
 
         self._lock_acquire()
 
@@ -743,17 +863,25 @@ class BlobStorage(SpecificationDecoratorBase):
                     data, serial_before, serial_after = load_result
                     orig_fn = self.fshelper.getBlobFilename(oid, serial_before)
                     new_fn = self.fshelper.getBlobFilename(oid, undo_serial)
-                orig = open(orig_fn, "r")
-                new = open(new_fn, "wb")
-                utils.cp(orig, new)
-                orig.close()
-                new.close()
+                with open(orig_fn, "rb") as orig:
+                    with open(new_fn, "wb") as new:
+                        utils.cp(orig, new)
                 self.dirty_oids.append((oid, undo_serial))
 
         finally:
             self._lock_release()
         return undo_serial, keys
 
+    def new_instance(self):
+        """Implementation of IMVCCStorage.new_instance.
+
+        This method causes all storage instances to be wrapped with
+        a blob storage wrapper.
+        """
+        base_dir = self.fshelper.base_dir
+        s = self.__storage.new_instance()
+        res = BlobStorage(base_dir, s)
+        return res
 
 copied = logging.getLogger('ZODB.blob.copied').debug
 def rename_or_copy_blob(f1, f2, chmod=True):
@@ -768,13 +896,9 @@ def rename_or_copy_blob(f1, f2, chmod=True):
         os.rename(f1, f2)
     except OSError:
         copied("Copied blob file %r to %r.", f1, f2)
-        file1 = open(f1, 'rb')
-        file2 = open(f2, 'wb')
-        try:
-            utils.cp(file1, file2)
-        finally:
-            file1.close()
-            file2.close()
+        with open(f1, 'rb') as file1:
+            with open(f2, 'wb') as file2:
+                utils.cp(file1, file2)
         remove_committed(f1)
     if chmod:
         os.chmod(f2, stat.S_IREAD)
@@ -793,6 +917,61 @@ if sys.platform == 'win32':
                 filename = os.path.join(dirpath, filename)
                 remove_committed(filename)
         shutil.rmtree(path)
+
+    link_or_copy = shutil.copy
 else:
     remove_committed = os.remove
     remove_committed_dir = shutil.rmtree
+    link_or_copy = os.link
+
+
+def find_global_Blob(module, class_):
+    if module == 'ZODB.blob' and class_ == 'Blob':
+        return Blob
+
+def is_blob_record(record):
+    """Check whether a database record is a blob record.
+
+    This is primarily intended to be used when copying data from one
+    storage to another.
+
+    """
+    if record and (b'ZODB.blob' in record):
+        unpickler = Unpickler(BytesIO(record))
+        unpickler.find_global = find_global_Blob
+
+        try:
+            return unpickler.load() is Blob
+        except (MemoryError, KeyboardInterrupt, SystemExit):
+            raise
+        except Exception:
+            pass
+
+    return False
+
+def copyTransactionsFromTo(source, destination):
+    for trans in source.iterator():
+        destination.tpc_begin(trans, trans.tid, trans.status)
+        for record in trans:
+            blobfilename = None
+            if is_blob_record(record.data):
+                try:
+                    blobfilename = source.loadBlob(record.oid, record.tid)
+                except POSKeyError:
+                    pass
+            if blobfilename is not None:
+                fd, name = tempfile.mkstemp(
+                    prefix='CTFT',
+                    suffix='.tmp', dir=destination.fshelper.temp_dir)
+                os.close(fd)
+                with open(blobfilename, 'rb') as sf:
+                    with open(name, 'wb') as df:
+                        utils.cp(sf, df)
+                destination.restoreBlob(record.oid, record.tid, record.data,
+                                 name, record.data_txn, trans)
+            else:
+                destination.restore(record.oid, record.tid, record.data,
+                             '', record.data_txn, trans)
+
+        destination.tpc_vote(trans)
+        destination.tpc_finish(trans)
