@@ -12,15 +12,21 @@
 #
 ##############################################################################
 """Broken object support
-
-$Id: broken.py 113732 2010-06-21 15:31:11Z ctheune $
 """
 
 import sys
+
 import persistent
+import zope.interface
+
+import ZODB.interfaces
+from ZODB._compat import IMPORT_MAPPING
+from ZODB._compat import NAME_MAPPING
+
 
 broken_cache = {}
 
+@zope.interface.implementer(ZODB.interfaces.IBroken)
 class Broken(object):
     """Broken object base class
 
@@ -29,7 +35,7 @@ class Broken(object):
 
        Broken objects don't really do much of anything, except hold their
        state.   The Broken class is used as a base class for creating
-       classes in leu of missing classes::
+       classes in lieu of missing classes::
 
          >>> Atall = type('Atall', (Broken, ), {'__module__': 'not.there'})
 
@@ -77,8 +83,10 @@ class Broken(object):
          >>> r[2]
          {'x': 1}
 
-         >>> import cPickle
-         >>> a2 = cPickle.loads(cPickle.dumps(a, 1))
+         >>> from ZODB._compat import dumps
+         >>> from ZODB._compat import loads
+         >>> from ZODB._compat import _protocol
+         >>> a2 = loads(dumps(a, _protocol))
          >>> a2
          <broken not.there.Atall instance>
          >>> a2.__Broken_newargs__
@@ -92,6 +100,7 @@ class Broken(object):
          >>> broken_cache.clear()
        """
 
+
     __Broken_state__ = __Broken_initargs__ = None
 
     __name__ = 'broken object'
@@ -135,6 +144,9 @@ def find_global(modulename, globalname,
          >>> find_global('sys', 'path') is sys.path
          True
 
+         >>> find_global('__builtin__', 'object') is object
+         True
+
        If an object can't be found, a broken class is returned::
 
          >>> broken = find_global('ZODB.not.there', 'atall')
@@ -177,6 +189,18 @@ def find_global(modulename, globalname,
 
          >>> broken_cache.clear()
        """
+
+    if (modulename, globalname) in NAME_MAPPING:
+        modulename, globalname = NAME_MAPPING[(modulename, globalname)]
+    if modulename in IMPORT_MAPPING:
+        modulename = IMPORT_MAPPING[modulename]
+
+    # short circuit common case:
+    try:
+        return getattr(sys.modules[modulename], globalname)
+    except (AttributeError, KeyError):
+        pass
+
     try:
         __import__(modulename)
     except ImportError:
