@@ -100,7 +100,8 @@ oid
     The following reference types are defined:
 
     'w'
-        Persistent weak reference.  The arguments consist of an oid.
+        Persistent weak reference.  The arguments consist of an oid
+        and optionally a database name.
 
     The following are planned for the future:
 
@@ -113,7 +114,7 @@ oid
         of a database name, an object id, and class meta data.
 
 The following legacy format is also supported.
-    
+
 [oid]
     A persistent weak reference
 
@@ -132,19 +133,18 @@ A number of legacyforms are defined:
 
 
 """
-
-import cPickle
-import cStringIO
 import logging
-
+import sys
 
 from persistent import Persistent
 from persistent.wref import WeakRefMarker, WeakRef
 from ZODB import broken
-from ZODB.broken import Broken
 from ZODB.POSException import InvalidObjectReference
+from ZODB._compat import Pickler, Unpickler, BytesIO, _protocol
+
+
+_oidtypes = bytes, type(None)
 
-_oidtypes = str, type(None)
 
 # Might to update or redo coptimizations to reflect weakrefs:
 # from ZODB.coptimizations import new_persistent_id
@@ -171,9 +171,12 @@ class ObjectWriter:
     _jar = None
 
     def __init__(self, obj=None):
-        self._file = cStringIO.StringIO()
-        self._p = cPickle.Pickler(self._file, 1)
-        self._p.persistent_id = self.persistent_id
+        self._file = BytesIO()
+        self._p = Pickler(self._file, _protocol)
+        if sys.version_info[0] < 3:
+            self._p.inst_persistent_id = self.persistent_id
+        else:
+            self._p.persistent_id = self.persistent_id
         self._stack = []
         if obj is not None:
             self._stack.append(obj)
@@ -186,12 +189,13 @@ class ObjectWriter:
 
         >>> from ZODB.tests.util import P
         >>> class DummyJar:
+        ...     xrefs = True
         ...     def new_oid(self):
         ...         return 42
         ...     def db(self):
         ...         return self
         ...     databases = {}
-        
+
         >>> jar = DummyJar()
         >>> class O:
         ...     _p_jar = jar
@@ -229,11 +233,13 @@ class ObjectWriter:
         If the jar doesn't match that of the writer, an error is raised:
 
         >>> bob._p_jar = DummyJar()
-        >>> writer.persistent_id(bob)   # doctest: +NORMALIZE_WHITESPACE
+        >>> writer.persistent_id(bob)
+        ... # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
         Traceback (most recent call last):
           ...
-        InvalidObjectReference: Attempt to store an object from a
-            foreign database connection
+        InvalidObjectReference:
+        ('Attempt to store an object from a foreign database connection',
+        <ZODB.serialize.DummyJar ...>, P(bob))
 
         Constructor arguments used by __new__(), as returned by
         __getnewargs__(), can affect memory allocation, but may also
@@ -280,7 +286,7 @@ class ObjectWriter:
             # Not persistent, pickle normally
             return None
 
-        if not (oid is None or isinstance(oid, str)):
+        if not (oid is None or isinstance(oid, bytes)):
             # Deserves a closer look:
 
             # Make sure it's not a descriptor
@@ -295,8 +301,8 @@ class ObjectWriter:
 
                 oid = obj.oid
                 if oid is None:
-                    obj = obj() # get the referenced object
-                    oid = obj._p_oid
+                    target = obj() # get the referenced object
+                    oid = target._p_oid
                     if oid is None:
                         # Here we are causing the object to be saved in
                         # the database. One could argue that we shouldn't
@@ -305,10 +311,16 @@ class ObjectWriter:
                         # assume that the object will be added eventually.
 
                         oid = self._jar.new_oid()
-                        obj._p_jar = self._jar
-                        obj._p_oid = oid
-                        self._stack.append(obj)
-                return ['w', (oid, )]
+                        target._p_jar = self._jar
+                        target._p_oid = oid
+                        self._stack.append(target)
+                    obj.oid = oid
+                    obj.dm = target._p_jar
+                    obj.database_name = obj.dm.db().database_name
+                if obj.dm is self._jar:
+                    return ['w', (oid, )]
+                else:
+                    return ['w', (oid, obj.database_name)]
 
 
         # Since we have an oid, we have either a persistent instance
@@ -322,8 +334,13 @@ class ObjectWriter:
             oid = obj._p_oid = self._jar.new_oid()
             obj._p_jar = self._jar
             self._stack.append(obj)
-            
+
         elif obj._p_jar is not self._jar:
+            if not self._jar.db().xrefs:
+                raise InvalidObjectReference(
+                    "Database %r doesn't allow implicit cross-database "
+                    "references" % self._jar.db().database_name,
+                    self._jar, obj)
 
             try:
                 otherdb = obj._p_jar.db()
@@ -334,14 +351,14 @@ class ObjectWriter:
             if self._jar.db().databases.get(database_name) is not otherdb:
                 raise InvalidObjectReference(
                     "Attempt to store an object from a foreign "
-                    "database connection"
+                    "database connection", self._jar, obj,
                     )
 
             if self._jar.get_connection(database_name) is not obj._p_jar:
                 raise InvalidObjectReference(
                     "Attempt to store a reference to an object from "
                     "a separate connection to the same database or "
-                    "multidatabase"
+                    "multidatabase", self._jar, obj,
                     )
 
             # OK, we have an object from another database.
@@ -350,9 +367,9 @@ class ObjectWriter:
             if obj._p_jar._implicitlyAdding(oid):
                 raise InvalidObjectReference(
                     "A new object is reachable from multiple databases. "
-                    "Won't try to guess which one was correct!"
+                    "Won't try to guess which one was correct!",
+                    self._jar, obj,
                     )
-                
 
         klass = type(obj)
         if hasattr(klass, '__getnewargs__'):
@@ -371,7 +388,7 @@ class ObjectWriter:
             return oid
 
         # Note that we never get here for persistent classes.
-        # We'll use driect refs for normal classes.
+        # We'll use direct refs for normal classes.
 
         if database_name is not None:
             return ['m', (database_name, oid, klass)]
@@ -407,7 +424,7 @@ class ObjectWriter:
         return self._dump(meta, obj.__getstate__())
 
     def _dump(self, classmeta, state):
-        # To reuse the existing cStringIO object, we must reset
+        # To reuse the existing BytesIO object, we must reset
         # the file position to 0 and truncate the file after the
         # new pickle is written.
         self._file.seek(0)
@@ -431,13 +448,15 @@ class NewObjectIterator:
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         if self._stack:
             elt = self._stack.pop()
             return elt
         else:
             raise StopIteration
 
+    next = __next__
+
 class ObjectReader:
 
     def __init__(self, conn=None, cache=None, factory=None):
@@ -449,8 +468,8 @@ class ObjectReader:
         return self._factory(self._conn, module, name)
 
     def _get_unpickler(self, pickle):
-        file = cStringIO.StringIO(pickle)
-        unpickler = cPickle.Unpickler(file)
+        file = BytesIO(pickle)
+        unpickler = Unpickler(file)
         unpickler.persistent_load = self._persistent_load
         factory = self._factory
         conn = self._conn
@@ -467,7 +486,7 @@ class ObjectReader:
     def _persistent_load(self, reference):
         if isinstance(reference, tuple):
             return self.load_persistent(*reference)
-        elif isinstance(reference, str):
+        elif isinstance(reference, (bytes, str)):
             return self.load_oid(reference)
         else:
             try:
@@ -482,6 +501,11 @@ class ObjectReader:
         # Quick instance reference.  We know all we need to know
         # to create the instance w/o hitting the db, so go for it!
 
+        if not isinstance(oid, bytes):
+            assert isinstance(oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            oid = oid.encode('ascii')
+
         obj = self._cache.get(oid, None)
         if obj is not None:
             return obj
@@ -489,7 +513,7 @@ class ObjectReader:
         if isinstance(klass, tuple):
             klass = self._get_class(*klass)
 
-        if issubclass(klass, Broken):
+        if issubclass(klass, broken.Broken):
             # We got a broken class. We might need to make it
             # PersistentBroken
             if not issubclass(klass, broken.PersistentBroken):
@@ -503,14 +527,7 @@ class ObjectReader:
             return self._conn.get(oid)
 
         # TODO: should be done by connection
-        obj._p_oid = oid
-        obj._p_jar = self._conn
-        # When an object is created, it is put in the UPTODATE
-        # state.  We must explicitly deactivate it to turn it into
-        # a ghost.
-        obj._p_changed = None
-
-        self._cache[oid] = obj
+        self._cache.new_ghost(oid, obj)
         return obj
 
     def load_multi_persistent(self, database_name, oid, klass):
@@ -522,15 +539,33 @@ class ObjectReader:
     loaders['m'] = load_multi_persistent
 
 
-    def load_persistent_weakref(self, oid):
+    def load_persistent_weakref(self, oid, database_name=None):
+        if not isinstance(oid, bytes):
+            assert isinstance(oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            oid = oid.encode('ascii')
         obj = WeakRef.__new__(WeakRef)
         obj.oid = oid
-        obj.dm = self._conn
+        if database_name is None:
+            obj.dm = self._conn
+        else:
+            obj.database_name = database_name
+            try:
+                obj.dm = self._conn.get_connection(database_name)
+            except KeyError:
+                # XXX Not sure what to do here.  It seems wrong to
+                # fail since this is a weak reference.  For now we'll
+                # just pretend that the target object has gone.
+                pass
         return obj
 
     loaders['w'] = load_persistent_weakref
 
     def load_oid(self, oid):
+        if not isinstance(oid, bytes):
+            assert isinstance(oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            oid = oid.encode('ascii')
         obj = self._cache.get(oid, None)
         if obj is not None:
             return obj
@@ -544,16 +579,6 @@ class ObjectReader:
 
     loaders['n'] = load_multi_oid
 
-    def _new_object(self, klass, args):
-        if not args and not myhasattr(klass, "__getnewargs__"):
-            obj = klass.__new__(klass)
-        else:
-            obj = klass(*args)
-            if not isinstance(klass, type):
-                obj.__dict__.clear()
-
-        return obj
-
     def getClassName(self, pickle):
         unpickler = self._get_unpickler(pickle)
         klass = unpickler.load()
@@ -582,7 +607,7 @@ class ObjectReader:
             # Definitely new style direct class reference
             args = ()
 
-        if issubclass(klass, Broken):
+        if issubclass(klass, broken.Broken):
             # We got a broken class. We might need to make it
             # PersistentBroken
             if not issubclass(klass, broken.PersistentBroken):
@@ -595,7 +620,7 @@ class ObjectReader:
         try:
             unpickler.load() # skip the class metadata
             return unpickler.load()
-        except EOFError, msg:
+        except EOFError as msg:
             log = logging.getLogger("ZODB.serialize")
             log.exception("Unpickling error: %r", pickle)
             raise
@@ -616,8 +641,8 @@ def referencesf(p, oids=None):
     """
 
     refs = []
-    u = cPickle.Unpickler(cStringIO.StringIO(p))
-    u.persistent_load = refs
+    u = Unpickler(BytesIO(p))
+    u.persistent_load = refs.append
     u.noload()
     u.noload()
 
@@ -626,22 +651,27 @@ def referencesf(p, oids=None):
 
     if oids is None:
         oids = []
-        
+
     for reference in refs:
         if isinstance(reference, tuple):
             oid = reference[0]
-        elif isinstance(reference, str):
+        elif isinstance(reference, (bytes, str)):
             oid = reference
         else:
             assert isinstance(reference, list)
             continue
 
+        if not isinstance(oid, bytes):
+            assert isinstance(oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            oid = oid.encode('ascii')
+
         oids.append(oid)
-    
+
     return oids
 
 oid_klass_loaders = {
-    'w': lambda oid: None,
+    'w': lambda oid, database_name=None: None,
     }
 
 def get_refs(a_pickle):
@@ -651,27 +681,32 @@ def get_refs(a_pickle):
     If the reference doesn't contain class information, then the
     klass information is None.
     """
-    
+
     refs = []
-    u = cPickle.Unpickler(cStringIO.StringIO(a_pickle))
-    u.persistent_load = refs
+    u = Unpickler(BytesIO(a_pickle))
+    u.persistent_load = refs.append
     u.noload()
     u.noload()
 
-    # Now we have a list of referencs.  Need to convert to list of
+    # Now we have a list of references.  Need to convert to list of
     # oids and class info:
 
     result = []
 
     for reference in refs:
         if isinstance(reference, tuple):
-            data = reference
-        elif isinstance(reference, str):
-            data = reference, None
+            oid, klass = reference
+        elif isinstance(reference, (bytes, str)):
+            data, klass = reference, None
         else:
             assert isinstance(reference, list)
             continue
 
-        result.append(data)
-    
+        if not isinstance(oid, bytes):
+            assert isinstance(oid, str)
+            # this happens on Python 3 when all bytes in the oid are < 0x80
+            oid = oid.encode('ascii')
+
+        result.append((oid, klass))
+
     return result
