@@ -18,20 +18,18 @@ semantics (which you can override), and it also provides a helper
 method _dostore() which performs a complete store transaction for a
 single object revision.
 """
-
+from __future__ import print_function
 import sys
 import time
-import types
-import unittest
-from cPickle import Pickler, Unpickler
-from cStringIO import StringIO
-
 import transaction
 
 from ZODB.utils import u64
 from ZODB.tests.MinPO import MinPO
+from ZODB._compat import Pickler, Unpickler, BytesIO, _protocol
+import ZODB.tests.util
 
-ZERO = '\0'*8
+
+ZERO = b'\0'*8
 
 def snooze():
     # In Windows, it's possible that two successive time.time() calls return
@@ -51,9 +49,12 @@ def _persistent_id(obj):
 
 def zodb_pickle(obj):
     """Create a pickle in the format expected by ZODB."""
-    f = StringIO()
-    p = Pickler(f, 1)
-    p.persistent_id = _persistent_id
+    f = BytesIO()
+    p = Pickler(f, _protocol)
+    if sys.version_info[0] < 3:
+        p.inst_persistent_id = _persistent_id
+    else:
+        p.persistent_id = _persistent_id
     klass = obj.__class__
     assert not hasattr(obj, '__getinitargs__'), "not ready for constructors"
     args = None
@@ -66,7 +67,7 @@ def zodb_pickle(obj):
 
     p.dump((klass, args))
     p.dump(state)
-    return f.getvalue(1)
+    return f.getvalue()
 
 def persistent_load(pid):
     # helper for zodb_unpickle
@@ -74,11 +75,11 @@ def persistent_load(pid):
 
 def zodb_unpickle(data):
     """Unpickle an object stored using the format expected by ZODB."""
-    f = StringIO(data)
+    f = BytesIO(data)
     u = Unpickler(f)
     u.persistent_load = persistent_load
     klass_info = u.load()
-    if isinstance(klass_info, types.TupleType):
+    if isinstance(klass_info, tuple):
         if isinstance(klass_info[0], type):
             # Unclear:  what is the second part of klass_info?
             klass, xxx = klass_info
@@ -96,7 +97,7 @@ def zodb_unpickle(data):
             try:
                 klass = ns[klassname]
             except KeyError:
-                print >> sys.stderr, "can't find %s in %r" % (klassname, ns)
+                print("can't find %s in %r" % (klassname, ns), file=sys.stderr)
         inst = klass()
     else:
         raise ValueError("expected class info: %s" % repr(klass_info))
@@ -119,13 +120,13 @@ def handle_all_serials(oid, *args):
     """
     d = {}
     for arg in args:
-        if isinstance(arg, types.StringType):
+        if isinstance(arg, bytes):
             d[oid] = arg
         elif arg is None:
             pass
         else:
             for oid, serial in arg:
-                if not isinstance(serial, types.StringType):
+                if not isinstance(serial, bytes):
                     raise serial # error from ZEO server
                 d[oid] = serial
     return d
@@ -142,14 +143,12 @@ def import_helper(name):
     return sys.modules[name]
 
 
-class StorageTestBase(unittest.TestCase):
+class StorageTestBase(ZODB.tests.util.TestCase):
 
     # It would be simpler if concrete tests didn't need to extend
     # setUp() and tearDown().
 
-    def setUp(self):
-        # You need to override this with a setUp that creates self._storage
-        self._storage = None
+    _storage = None
 
     def _close(self):
         # You should override this if closing your storage requires additional
@@ -159,15 +158,15 @@ class StorageTestBase(unittest.TestCase):
 
     def tearDown(self):
         self._close()
+        ZODB.tests.util.TestCase.tearDown(self)
 
-    def _dostore(self, oid=None, revid=None, data=None, version=None,
+    def _dostore(self, oid=None, revid=None, data=None,
                  already_pickled=0, user=None, description=None):
         """Do a complete storage transaction.  The defaults are:
 
          - oid=None, ask the storage for a new oid
          - revid=None, use a revid of ZERO
          - data=None, pickle up some arbitrary data (the integer 7)
-         - version=None, use the empty string version
 
         Returns the object's new revision id.
         """
@@ -177,12 +176,10 @@ class StorageTestBase(unittest.TestCase):
             revid = ZERO
         if data is None:
             data = MinPO(7)
-        if type(data) == types.IntType:
+        if type(data) == int:
             data = MinPO(data)
         if not already_pickled:
             data = zodb_pickle(data)
-        if version is None:
-            version = ''
         # Begin the transaction
         t = transaction.Transaction()
         if user is not None:
@@ -192,7 +189,7 @@ class StorageTestBase(unittest.TestCase):
         try:
             self._storage.tpc_begin(t)
             # Store an object
-            r1 = self._storage.store(oid, revid, data, version, t)
+            r1 = self._storage.store(oid, revid, data, '', t)
             # Finish the transaction
             r2 = self._storage.tpc_vote(t)
             revid = handle_serials(oid, r1, r2)
@@ -202,9 +199,9 @@ class StorageTestBase(unittest.TestCase):
             raise
         return revid
 
-    def _dostoreNP(self, oid=None, revid=None, data=None, version=None,
+    def _dostoreNP(self, oid=None, revid=None, data=None,
                    user=None, description=None):
-        return self._dostore(oid, revid, data, version, 1, user, description)
+        return self._dostore(oid, revid, data, 1, user, description)
 
     # The following methods depend on optional storage features.
 
@@ -214,29 +211,13 @@ class StorageTestBase(unittest.TestCase):
         t = transaction.Transaction()
         t.note(note or "undo")
         self._storage.tpc_begin(t)
-        tid, oids = self._storage.undo(tid, t)
-        self._storage.tpc_vote(t)
+        undo_result = self._storage.undo(tid, t)
+        vote_result = self._storage.tpc_vote(t)
         self._storage.tpc_finish(t)
         if expected_oids is not None:
+            oids = list(undo_result[1]) if undo_result else []
+            oids.extend(oid for (oid, _) in vote_result or ())
             self.assertEqual(len(oids), len(expected_oids), repr(oids))
             for oid in expected_oids:
-                self.assert_(oid in oids)
+                self.assertTrue(oid in oids)
         return self._storage.lastTransaction()
-
-    def _commitVersion(self, src, dst):
-        t = transaction.Transaction()
-        t.note("commit %r to %r" % (src, dst))
-        self._storage.tpc_begin(t)
-        tid, oids = self._storage.commitVersion(src, dst, t)
-        self._storage.tpc_vote(t)
-        self._storage.tpc_finish(t)
-        return oids
-
-    def _abortVersion(self, ver):
-        t = transaction.Transaction()
-        t.note("abort %r" % ver)
-        self._storage.tpc_begin(t)
-        tid, oids = self._storage.abortVersion(ver, t)
-        self._storage.tpc_vote(t)
-        self._storage.tpc_finish(t)
-        return oids
