@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -11,14 +11,20 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-import os, unittest
+import doctest
+import os
+if os.environ.get('USE_ZOPE_TESTING_DOCTEST'):
+    from zope.testing import doctest
+import unittest
 import transaction
 import ZODB.FileStorage
+import ZODB.tests.hexstorage
 import ZODB.tests.testblob
 import ZODB.tests.util
 import zope.testing.setupstack
 from ZODB import POSException
 from ZODB import DB
+from ZODB.fsIndex import fsIndex
 
 from ZODB.tests import StorageTestBase, BasicStorage, TransactionalUndoStorage
 from ZODB.tests import PackableStorage, Synchronization, ConflictResolution
@@ -26,6 +32,8 @@ from ZODB.tests import HistoryStorage, IteratorStorage, Corruption
 from ZODB.tests import RevisionStorage, PersistentStorage, MTStorage
 from ZODB.tests import ReadOnlyStorage, RecoveryStorage
 from ZODB.tests.StorageTestBase import MinPO, zodb_pickle
+from ZODB._compat import dump, dumps, _protocol
+
 
 class FileStorageTests(
     StorageTestBase.StorageTestBase,
@@ -69,7 +77,6 @@ class FileStorageTests(
             self.fail("expect long user field to raise error")
 
     def check_use_fsIndex(self):
-        from ZODB.fsIndex import fsIndex
 
         self.assertEqual(self._storage._index.__class__, fsIndex)
 
@@ -78,21 +85,14 @@ class FileStorageTests(
     def convert_index_to_dict(self):
         # Convert the index in the current .index file to a Python dict.
         # Return the index originally found.
-        import cPickle as pickle
-
-        f = open('FileStorageTests.fs.index', 'r+b')
-        p = pickle.Unpickler(f)
-        data = p.load()
+        data = fsIndex.load('FileStorageTests.fs.index')
         index = data['index']
 
         newindex = dict(index)
         data['index'] = newindex
 
-        f.seek(0)
-        f.truncate()
-        p = pickle.Pickler(f, 1)
-        p.dump(data)
-        f.close()
+        with open('FileStorageTests.fs.index', 'wb') as fp:
+            dump(data, fp, _protocol)
         return index
 
     def check_conversion_to_fsIndex(self, read_only=False):
@@ -108,13 +108,13 @@ class FileStorageTests(
 
         # Convert it to a dict.
         old_index = self.convert_index_to_dict()
-        self.assert_(isinstance(old_index, fsIndex))
+        self.assertTrue(isinstance(old_index, fsIndex))
         new_index = self.convert_index_to_dict()
-        self.assert_(isinstance(new_index, dict))
+        self.assertTrue(isinstance(new_index, dict))
 
         # Verify it's converted to fsIndex in memory upon open.
         self.open(read_only=read_only)
-        self.assert_(isinstance(self._storage._index, fsIndex))
+        self.assertTrue(isinstance(self._storage._index, fsIndex))
 
         # Verify it has the right content.
         newindex_as_dict = dict(self._storage._index)
@@ -124,9 +124,9 @@ class FileStorageTests(
         self._storage.close()
         current_index = self.convert_index_to_dict()
         if read_only:
-            self.assert_(isinstance(current_index, dict))
+            self.assertTrue(isinstance(current_index, dict))
         else:
-            self.assert_(isinstance(current_index, fsIndex))
+            self.assertTrue(isinstance(current_index, fsIndex))
 
     def check_conversion_to_fsIndex_readonly(self):
         # Same thing, but the disk .index should continue to hold a
@@ -156,8 +156,8 @@ class FileStorageTests(
 
         # Verify it's converted to fsIndex in memory upon open.
         self.open()
-        self.assert_(isinstance(self._storage._index, fsIndex))
-        self.assert_(isinstance(self._storage._index._data, OOBTree))
+        self.assertTrue(isinstance(self._storage._index, fsIndex))
+        self.assertTrue(isinstance(self._storage._index._data, OOBTree))
 
         # Verify it has the right content.
         new_data_dict = dict(self._storage._index._data)
@@ -182,10 +182,10 @@ class FileStorageTests(
         # of the largest oid in use.
         t = transaction.Transaction()
         self._storage.tpc_begin(t)
-        giant_oid = '\xee' * 8
+        giant_oid = b'\xee' * 8
         # Store an object.
         # oid, serial, data, version, transaction
-        r1 = self._storage.store(giant_oid, '\0'*8, 'data', '', t)
+        r1 = self._storage.store(giant_oid, b'\0'*8, b'data', b'', t)
         # Finish the transaction.
         r2 = self._storage.tpc_vote(t)
         self._storage.tpc_finish(t)
@@ -199,10 +199,10 @@ class FileStorageTests(
         # ZRS recovery, use the .restore() method, this is plain critical.
         t = transaction.Transaction()
         self._storage.tpc_begin(t)
-        giant_oid = '\xee' * 8
+        giant_oid = b'\xee' * 8
         # Store an object.
         # oid, serial, data, version, prev_txn, transaction
-        r1 = self._storage.restore(giant_oid, '\0'*8, 'data', '', None, t)
+        r1 = self._storage.restore(giant_oid, b'\0'*8, b'data', b'', None, t)
         # Finish the transaction.
         r2 = self._storage.tpc_vote(t)
         self._storage.tpc_finish(t)
@@ -234,27 +234,26 @@ class FileStorageTests(
         self.open()
 
         # Open .fs directly, and damage content.
-        f = open('FileStorageTests.fs', 'r+b')
-        f.seek(0, 2)
-        pos2 = f.tell() - 8
-        f.seek(pos2)
-        tlen2 = U64(f.read(8))  # length-8 of the last transaction
-        pos1 = pos2 - tlen2 + 8 # skip over the tid at the start
-        f.seek(pos1)
-        tlen1 = U64(f.read(8))  # should be redundant length-8
-        self.assertEqual(tlen1, tlen2)  # verify that it is redundant
+        with open('FileStorageTests.fs', 'r+b') as f:
+            f.seek(0, 2)
+            pos2 = f.tell() - 8
+            f.seek(pos2)
+            tlen2 = U64(f.read(8))  # length-8 of the last transaction
+            pos1 = pos2 - tlen2 + 8 # skip over the tid at the start
+            f.seek(pos1)
+            tlen1 = U64(f.read(8))  # should be redundant length-8
+            self.assertEqual(tlen1, tlen2)  # verify that it is redundant
 
-        # Now damage the second copy.
-        f.seek(pos2)
-        f.write(p64(tlen2 - 1))
-        f.close()
+            # Now damage the second copy.
+            f.seek(pos2)
+            f.write(p64(tlen2 - 1))
 
         # Try to pack.  This used to yield
         #     NameError: global name 's' is not defined
         try:
             self._storage.pack(time.time(), referencesf)
-        except CorruptedError, detail:
-            self.assert_("redundant transaction length does not match "
+        except CorruptedError as detail:
+            self.assertTrue("redundant transaction length does not match "
                          "initial transaction length" in str(detail))
         else:
             self.fail("expected CorruptedError")
@@ -274,25 +273,41 @@ class FileStorageTests(
         self.open()
 
         key = None
-        for x in ('\000', '\001', '\002'):
+        for x in (b'\000', b'\001', b'\002'):
             oid, tid, data, next_oid = self._storage.record_iternext(key)
-            self.assertEqual(oid, ('\000' * 7) + x)
+            self.assertEqual(oid, (b'\000' * 7) + x)
             key = next_oid
             expected_data, expected_tid = self._storage.load(oid, '')
             self.assertEqual(expected_data, data)
             self.assertEqual(expected_tid, tid)
-            if x == '\002':
+            if x == b'\002':
                 self.assertEqual(next_oid, None)
             else:
                 self.assertNotEqual(next_oid, None)
 
+class FileStorageHexTests(FileStorageTests):
+
+    def open(self, **kwargs):
+        self._storage = ZODB.tests.hexstorage.HexStorage(
+            ZODB.FileStorage.FileStorage('FileStorageTests.fs',**kwargs))
+
+
 class FileStorageTestsWithBlobsEnabled(FileStorageTests):
 
     def open(self, **kwargs):
         if 'blob_dir' not in kwargs:
             kwargs = kwargs.copy()
             kwargs['blob_dir'] = 'blobs'
-        return FileStorageTests.open(self, **kwargs)
+        FileStorageTests.open(self, **kwargs)
+
+class FileStorageHexTestsWithBlobsEnabled(FileStorageTests):
+
+    def open(self, **kwargs):
+        if 'blob_dir' not in kwargs:
+            kwargs = kwargs.copy()
+            kwargs['blob_dir'] = 'blobs'
+        FileStorageTests.open(self, **kwargs)
+        self._storage = ZODB.tests.hexstorage.HexStorage(self._storage)
 
 class FileStorageRecoveryTest(
     StorageTestBase.StorageTestBase,
@@ -311,6 +326,15 @@ class FileStorageRecoveryTest(
     def new_dest(self):
         return ZODB.FileStorage.FileStorage('Dest.fs')
 
+class FileStorageHexRecoveryTest(FileStorageRecoveryTest):
+
+    def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
+        self._storage = ZODB.tests.hexstorage.HexStorage(
+            ZODB.FileStorage.FileStorage("Source.fs", create=True))
+        self._dst = ZODB.tests.hexstorage.HexStorage(
+            ZODB.FileStorage.FileStorage("Dest.fs", create=True))
+
 
 class FileStorageNoRestore(ZODB.FileStorage.FileStorage):
 
@@ -344,36 +368,38 @@ class AnalyzeDotPyTest(StorageTestBase.StorageTestBase):
         self._storage = ZODB.FileStorage.FileStorage("Source.fs", create=True)
 
     def checkanalyze(self):
-        import new, sys, pickle
+        import types, sys
         from BTrees.OOBTree import OOBTree
         from ZODB.scripts import analyze
-        
+
         # Set up a module to act as a broken import
         module_name = 'brokenmodule'
-        module = new.module(module_name)
+        module = types.ModuleType(module_name)
         sys.modules[module_name] = module
-        
+
         class Broken(MinPO):
             __module__ = module_name
         module.Broken = Broken
-        
+
         oids = [[self._storage.new_oid(), None] for i in range(3)]
         for i in range(2):
             t = transaction.Transaction()
             self._storage.tpc_begin(t)
-            
+
             # sometimes data is in this format
             j = 0
             oid, revid = oids[j]
-            serial = self._storage.store(oid, revid, pickle.dumps(OOBTree, 1), "", t) 
+            serial = self._storage.store(
+                oid, revid, dumps(OOBTree, _protocol), "", t)
             oids[j][1] = serial
-            
+
             # and it could be from a broken module
             j = 1
             oid, revid = oids[j]
-            serial = self._storage.store(oid, revid, pickle.dumps(Broken, 1), "", t) 
+            serial = self._storage.store(
+                oid, revid, dumps(Broken, _protocol), "", t)
             oids[j][1] = serial
-            
+
             # but mostly it looks like this
             j = 2
             o = MinPO(j)
@@ -383,30 +409,30 @@ class AnalyzeDotPyTest(StorageTestBase.StorageTestBase):
 
             self._storage.tpc_vote(t)
             self._storage.tpc_finish(t)
-            
+
         # now break the import of the Broken class
         del sys.modules[module_name]
-        
+
         # from ZODB.scripts.analyze.analyze
         fsi = self._storage.iterator()
         rep = analyze.Report()
         for txn in fsi:
             analyze.analyze_trans(rep, txn)
-        
+
         # from ZODB.scripts.analyze.report
-        typemap = rep.TYPEMAP.keys()
-        typemap.sort()
+        typemap = sorted(rep.TYPEMAP.keys())
         cumpct = 0.0
         for t in typemap:
             pct = rep.TYPESIZE[t] * 100.0 / rep.DBYTES
             cumpct += pct
-       
-        self.assertAlmostEqual(cumpct, 100.0, 0, "Failed to analyze some records")
+
+        self.assertAlmostEqual(cumpct, 100.0, 0,
+                               "Failed to analyze some records")
 
 # Raise an exception if the tids in FileStorage fs aren't
 # strictly increasing.
 def checkIncreasingTids(fs):
-    lasttid = '\0' * 8
+    lasttid = b'\0' * 8
     for txn in fs.iterator():
         if lasttid >= txn.tid:
             raise ValueError("tids out of order %r >= %r" % (lasttid, txn.tid))
@@ -572,22 +598,22 @@ def deal_with_finish_failures():
     >>> import zope.testing.loggingsupport
     >>> handler = zope.testing.loggingsupport.InstalledHandler(
     ...     'ZODB.FileStorage')
-    >>> transaction.commit()
+    >>> transaction.commit() # doctest: +ELLIPSIS
     Traceback (most recent call last):
     ...
-    TypeError: <lambda>() takes no arguments (1 given)
+    TypeError: <lambda>() takes ...
 
 
-    >>> print handler
+    >>> print(handler)
     ZODB.FileStorage CRITICAL
       Failure in _finish. Closing.
 
     >>> handler.uninstall()
 
-    >>> fs.load('\0'*8, '')
+    >>> fs.load(b'\0'*8, b'') # doctest: +ELLIPSIS
     Traceback (most recent call last):
     ...
-    ValueError: I/O operation on closed file
+    ValueError: ...
 
     >>> db.close()
     >>> fs = ZODB.FileStorage.FileStorage('data.fs')
@@ -611,7 +637,8 @@ def pack_with_open_blob_files():
     >>> import ZODB.blob
     >>> conn1.root()[1] = ZODB.blob.Blob()
     >>> conn1.add(conn1.root()[1])
-    >>> conn1.root()[1].open('w').write('some data')
+    >>> with conn1.root()[1].open('w') as file:
+    ...     _ = file.write(b'some data')
     >>> tm1.commit()
 
     >>> tm2 = transaction.TransactionManager()
@@ -619,7 +646,8 @@ def pack_with_open_blob_files():
     >>> f = conn1.root()[1].open()
     >>> conn1.root()[2] = ZODB.blob.Blob()
     >>> conn1.add(conn1.root()[2])
-    >>> conn1.root()[2].open('w').write('some more data')
+    >>> with conn1.root()[2].open('w') as file:
+    ...     _ = file.write(b'some more data')
 
     >>> db.pack()
     >>> f.read()
@@ -628,24 +656,27 @@ def pack_with_open_blob_files():
 
     >>> tm1.commit()
     >>> conn2.sync()
-    >>> conn2.root()[2].open().read()
+    >>> with conn2.root()[2].open() as fp: fp.read()
     'some more data'
 
     >>> db.close()
     """
 
 def test_suite():
-    from zope.testing import doctest
-
     suite = unittest.TestSuite()
-    for klass in [FileStorageTests, Corruption.FileStorageCorruptTests,
-                  FileStorageRecoveryTest, FileStorageNoRestoreRecoveryTest,
-                  FileStorageTestsWithBlobsEnabled, AnalyzeDotPyTest,
-                  ]:
+    for klass in [
+        FileStorageTests, FileStorageHexTests,
+        Corruption.FileStorageCorruptTests,
+        FileStorageRecoveryTest, FileStorageHexRecoveryTest,
+        FileStorageNoRestoreRecoveryTest,
+        FileStorageTestsWithBlobsEnabled, FileStorageHexTestsWithBlobsEnabled,
+        AnalyzeDotPyTest,
+        ]:
         suite.addTest(unittest.makeSuite(klass, "check"))
     suite.addTest(doctest.DocTestSuite(
         setUp=zope.testing.setupstack.setUpDirectory,
-        tearDown=zope.testing.setupstack.tearDown))
+        tearDown=zope.testing.setupstack.tearDown,
+        checker=ZODB.tests.util.checker))
     suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
         'BlobFileStorage',
         lambda name, blob_dir:
@@ -653,9 +684,18 @@ def test_suite():
         test_blob_storage_recovery=True,
         test_packing=True,
         ))
+    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
+        'BlobFileHexStorage',
+        lambda name, blob_dir:
+        ZODB.tests.hexstorage.HexStorage(
+            ZODB.FileStorage.FileStorage('%s.fs' % name, blob_dir=blob_dir)),
+        test_blob_storage_recovery=True,
+        test_packing=True,
+        ))
     suite.addTest(PackableStorage.IExternalGC_suite(
         lambda : ZODB.FileStorage.FileStorage(
             'data.fs', blob_dir='blobs', pack_gc=False)))
+    suite.layer = ZODB.tests.util.MininalTestLayer('testFileStorage')
     return suite
 
 if __name__=='__main__':
