@@ -11,17 +11,13 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-
 r"""
 fsdump test
 ===========
 
-Let's get a temp file path to work with first.
+Let's get a path to work with first.
 
->>> import tempfile
->>> path = tempfile.mktemp('.fs', 'Data')
->>> print 'path:', path #doctest: +ELLIPSIS
-path: ...Data...fs
+>>> path = 'Data.fs'
 
 More imports.
 
@@ -43,9 +39,9 @@ Create a root object and try again:
 
 >>> db = ZODB.DB(st) # yes, that creates a root object!
 >>> fsdump(path) #doctest: +ELLIPSIS
-Trans #00000 tid=... time=... offset=52
+Trans #00000 tid=... time=... offset=<OFFSET>
     status=' ' user='' description='initial database creation'
-  data #00000 oid=0000000000000000 size=66 class=persistent.mapping.PersistentMapping
+  data #00000 oid=0000000000000000 size=<SIZE> class=persistent.mapping.PersistentMapping
 
 Now we see first transaction with root object.
 
@@ -56,23 +52,41 @@ Let's add a BTree:
 >>> txn.get().note('added an OOBTree')
 >>> txn.get().commit()
 >>> fsdump(path) #doctest: +ELLIPSIS
-Trans #00000 tid=... time=... offset=52
+Trans #00000 tid=... time=... offset=<OFFSET>
     status=' ' user='' description='initial database creation'
-  data #00000 oid=0000000000000000 size=66 class=persistent.mapping.PersistentMapping
-Trans #00001 tid=... time=... offset=207
+  data #00000 oid=0000000000000000 size=<SIZE> class=persistent.mapping.PersistentMapping
+Trans #00001 tid=... time=... offset=<OFFSET>
     status=' ' user='' description='added an OOBTree'
-  data #00000 oid=0000000000000000 size=113 class=persistent.mapping.PersistentMapping
-  data #00001 oid=0000000000000001 size=29 class=BTrees.OOBTree.OOBTree
+  data #00000 oid=0000000000000000 size=<SIZE> class=persistent.mapping.PersistentMapping
+  data #00001 oid=0000000000000001 size=<SIZE> class=BTrees.OOBTree.OOBTree
 
 Now we see two transactions and two changed objects.
 
 Clean up.
 
 >>> st.close()
->>> st.cleanup() # remove .fs, .index, etc
 """
 
-from zope.testing import doctest
+import re
+import doctest
+import zope.testing.setupstack
+import ZODB.tests.util
+from zope.testing import renormalizing
+
+checker = renormalizing.RENormalizing([
+    # Normalizing this makes diffs easier to read
+    (re.compile(r'\btid=[0-9a-f]+\b'), 'tid=...'),
+    (re.compile(r'\b\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d\.\d+\b'), '...'),
+    # Python 3 produces larger pickles, even when we use zodbpickle :(
+    # this changes all the offsets and sizes
+    (re.compile(r'\bsize=[0-9]+\b'), 'size=<SIZE>'),
+    (re.compile(r'\offset=[0-9]+\b'), 'offset=<OFFSET>'),
+])
+
 
 def test_suite():
-    return doctest.DocTestSuite()
+    return doctest.DocTestSuite(
+        setUp=zope.testing.setupstack.setUpDirectory,
+        tearDown=zope.testing.setupstack.tearDown,
+        optionflags=doctest.REPORT_NDIFF,
+        checker=ZODB.tests.util.checker + checker)
