@@ -11,17 +11,13 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-
 r"""
 fsoids test, of the workhorse fsoids.Trace class
 ================================================
 
-Let's get a temp file path to work with first.
+Let's get a path to work with first.
 
->>> import tempfile
->>> path = tempfile.mktemp('.fs', 'Data')
->>> print 'path:', path #doctest: +ELLIPSIS
-path: ...Data...fs
+>>> path = 'Data.fs'
 
 More imports.
 
@@ -60,10 +56,10 @@ Create a root object and try again:
 >>> t.register_oids(0, 1)
 >>> t.run(); t.report() #doctest: +ELLIPSIS
 oid 0x00 persistent.mapping.PersistentMapping 1 revision
-    tid 0x... offset=4 ...
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='initial database creation'
-        new revision persistent.mapping.PersistentMapping at 52
+        new revision persistent.mapping.PersistentMapping at <OFFSET>
 oid 0x01 <unknown> 0 revisions
     this oid was not defined (no data record for it found)
 
@@ -86,21 +82,21 @@ Let's add a BTree and try again:
 >>> t.register_oids(0, 1)
 >>> t.run(); t.report() #doctest: +ELLIPSIS
 oid 0x00 persistent.mapping.PersistentMapping 2 revisions
-    tid 0x... offset=4 ...
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='initial database creation'
-        new revision persistent.mapping.PersistentMapping at 52
-    tid 0x... offset=168 ...
+        new revision persistent.mapping.PersistentMapping at <OFFSET>
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='added an OOBTree'
-        new revision persistent.mapping.PersistentMapping at 207
-        references 0x01 BTrees.OOBTree.OOBTree at 207
+        new revision persistent.mapping.PersistentMapping at <OFFSET>
+        references 0x01 BTrees.OOBTree.OOBTree at <OFFSET>
 oid 0x01 BTrees.OOBTree.OOBTree 1 revision
-    tid 0x... offset=168 ...
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='added an OOBTree'
-        new revision BTrees.OOBTree.OOBTree at 362
-        referenced by 0x00 persistent.mapping.PersistentMapping at 207
+        new revision BTrees.OOBTree.OOBTree at <OFFSET>
+        referenced by 0x00 persistent.mapping.PersistentMapping at <OFFSET>
 
 So there are two revisions of oid 0 now, and the second references oid 1.
 
@@ -114,30 +110,30 @@ One more, storing a reference in the BTree back to the root object:
 >>> t.register_oids(0, 1, 2)
 >>> t.run(); t.report() #doctest: +ELLIPSIS
 oid 0x00 persistent.mapping.PersistentMapping 2 revisions
-    tid 0x... offset=4 ...
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='initial database creation'
-        new revision persistent.mapping.PersistentMapping at 52
-    tid 0x... offset=168 ...
+        new revision persistent.mapping.PersistentMapping at <OFFSET>
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='added an OOBTree'
-        new revision persistent.mapping.PersistentMapping at 207
-        references 0x01 BTrees.OOBTree.OOBTree at 207
-    tid 0x... offset=441 ...
+        new revision persistent.mapping.PersistentMapping at <OFFSET>
+        references 0x01 BTrees.OOBTree.OOBTree at <OFFSET>
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='circling back to the root'
-        referenced by 0x01 BTrees.OOBTree.OOBTree at 489
+        referenced by 0x01 BTrees.OOBTree.OOBTree at <OFFSET>
 oid 0x01 BTrees.OOBTree.OOBTree 2 revisions
-    tid 0x... offset=168 ...
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='added an OOBTree'
-        new revision BTrees.OOBTree.OOBTree at 362
-        referenced by 0x00 persistent.mapping.PersistentMapping at 207
-    tid 0x... offset=441 ...
+        new revision BTrees.OOBTree.OOBTree at <OFFSET>
+        referenced by 0x00 persistent.mapping.PersistentMapping at <OFFSET>
+    tid 0x... offset=<OFFSET> ...
         tid user=''
         tid description='circling back to the root'
-        new revision BTrees.OOBTree.OOBTree at 489
-        references 0x00 persistent.mapping.PersistentMapping at 489
+        new revision BTrees.OOBTree.OOBTree at <OFFSET>
+        references 0x00 persistent.mapping.PersistentMapping at <OFFSET>
 oid 0x02 <unknown> 0 revisions
     this oid was not defined (no data record for it found)
 
@@ -171,7 +167,28 @@ Clean up.
 >>> st.cleanup() # remove .fs, .index, etc
 """
 
-from zope.testing import doctest
+import doctest
+import re
+
+from zope.testing import renormalizing
+
+from .util import checker as util_checker
+from .util import setUp
+from .util import tearDown
+
+checker = renormalizing.RENormalizing([
+    # Normalizing this makes diffs easier to read
+    (re.compile(r'\btid 0x[0-9a-f]+\b'), 'tid 0x...'),
+    (re.compile(r'\b\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d\.\d+\b'), '...'),
+    # Python 3 produces larger pickles, even when we use zodbpickle :(
+    # this changes all the offsets and sizes
+    (re.compile(r'\boffset=[0-9]+\b'), 'offset=<OFFSET>'),
+    (re.compile(r'\bat [0-9]+'), 'at <OFFSET>'),
+])
+
 
 def test_suite():
-    return doctest.DocTestSuite()
+    return doctest.DocTestSuite(setUp=setUp,
+                                tearDown=tearDown,
+                                checker=util_checker + checker,
+                                optionflags=doctest.REPORT_NDIFF)
