@@ -11,21 +11,22 @@
 # FOR A PARTICULAR PURPOSE
 #
 ##############################################################################
-
+import os
+import struct
 import sys
 import time
-import struct
-from struct import pack, unpack
-from binascii import hexlify, unhexlify
-import cPickle as pickle
-from cStringIO import StringIO
-import weakref
 import warnings
+from binascii import hexlify, unhexlify
+from struct import pack, unpack
 from tempfile import mkstemp
-import os
 
 from persistent.TimeStamp import TimeStamp
 
+from ZODB._compat import Unpickler
+from ZODB._compat import BytesIO
+from ZODB._compat import ascii_bytes
+
+
 __all__ = ['z64',
            'p64',
            'u64',
@@ -67,7 +68,37 @@ def deprecated38(msg):
     warnings.warn("This will be removed in ZODB 3.8:\n%s" % msg,
                   DeprecationWarning, stacklevel=3)
 
-z64 = '\0'*8
+
+if sys.version_info[0] < 3:
+    def as_bytes(obj):
+        "Convert obj into bytes"
+        return str(obj)
+
+    def as_text(bytes):
+        "Convert bytes into string"
+        return bytes
+
+    # Convert an element of a bytes object into an int
+    byte_ord = ord
+    byte_chr = chr
+
+else:
+    def as_bytes(obj):
+        if isinstance(obj, bytes):
+            # invoking str on a bytes object gives its repr()
+            return obj
+        return str(obj).encode("ascii")
+
+    def as_text(bytes):
+        return bytes.decode("ascii")
+
+    def byte_ord(byte):
+        return byte # elements of bytes are already ints
+
+    def byte_chr(int):
+        return bytes((int,))
+
+z64 = b'\0' * 8
 
 assert sys.hexversion >= 0x02030000
 
@@ -118,27 +149,28 @@ def newTid(old):
     ts = TimeStamp(*time.gmtime(t)[:5]+(t%60,))
     if old is not None:
         ts = ts.laterThan(TimeStamp(old))
-    return `ts`
+    return ts.raw()
 
 
 def oid_repr(oid):
-    if isinstance(oid, str) and len(oid) == 8:
+    if isinstance(oid, bytes) and len(oid) == 8:
         # Convert to hex and strip leading zeroes.
-        as_hex = hexlify(oid).lstrip('0')
+        as_hex = hexlify(oid).lstrip(b'0')
         # Ensure two characters per input byte.
         if len(as_hex) & 1:
-            as_hex = '0' + as_hex
-        elif as_hex == '':
-            as_hex = '00'
-        return '0x' + as_hex
+            as_hex = b'0' + as_hex
+        elif as_hex == b'':
+            as_hex = b'00'
+        return '0x' + as_hex.decode()
     else:
         return repr(oid)
 
 def repr_to_oid(repr):
-    if repr.startswith("0x"):
+    repr = ascii_bytes(repr)
+    if repr.startswith(b"0x"):
         repr = repr[2:]
     as_bin = unhexlify(repr)
-    as_bin = "\x00"*(8-len(as_bin)) + as_bin
+    as_bin = b"\x00"*(8-len(as_bin)) + as_bin
     return as_bin
 
 serial_repr = oid_repr
@@ -146,7 +178,7 @@ tid_repr = serial_repr
 
 # For example, produce
 #     '0x03441422948b4399 2002-04-14 20:50:34.815000'
-# for 8-byte string tid '\x03D\x14"\x94\x8bC\x99'.
+# for 8-byte string tid b'\x03D\x14"\x94\x8bC\x99'.
 def readable_tid_repr(tid):
     result = tid_repr(tid)
     if isinstance(tid, str) and len(tid) == 8:
@@ -180,12 +212,18 @@ def positive_id(obj):
 # for what serialize.py calls formats 5 and 6.
 
 def get_pickle_metadata(data):
+    # Returns a 2-tuple of strings.
+
     # ZODB's data records contain two pickles.  The first is the class
     # of the object, the second is the object.  We're only trying to
     # pick apart the first here, to extract the module and class names.
-    if data.startswith('(c'):   # pickle MARK GLOBAL opcode sequence
+    if data[0] in (0x80,    # Py3k indexes bytes -> int
+                   b'\x80'  # Python2 indexes bytes -> bytes
+                  ): # protocol marker, protocol > 1
+        data = data[2:]
+    if data.startswith(b'(c'):   # pickle MARK GLOBAL opcode sequence
         global_prefix = 2
-    elif data.startswith('c'):  # pickle GLOBAL opcode
+    elif data.startswith(b'c'):  # pickle GLOBAL opcode
         global_prefix = 1
     else:
         global_prefix = 0
@@ -196,16 +234,16 @@ def get_pickle_metadata(data):
         # load the class.  Just break open the pickle and get the
         # module and class from it.  The module and class names are given by
         # newline-terminated strings following the GLOBAL opcode.
-        modname, classname, rest = data.split('\n', 2)
+        modname, classname, rest = data.split(b'\n', 2)
         modname = modname[global_prefix:]   # strip GLOBAL opcode
-        return modname, classname
+        return modname.decode(), classname.decode()
 
     # Else there are a bunch of other possible formats.
-    f = StringIO(data)
-    u = pickle.Unpickler(f)
+    f = BytesIO(data)
+    u = Unpickler(f)
     try:
         class_info = u.load()
-    except Exception, err:
+    except Exception as err:
         return '', ''
     if isinstance(class_info, tuple):
         if isinstance(class_info[0], tuple):
@@ -220,28 +258,29 @@ def get_pickle_metadata(data):
         classname = ''
     return modname, classname
 
-def mktemp(dir=None):
+def mktemp(dir=None, prefix='tmp'):
     """Create a temp file, known by name, in a semi-secure manner."""
-    handle, filename = mkstemp(dir=dir)
+    handle, filename = mkstemp(dir=dir, prefix=prefix)
     os.close(handle)
     return filename
 
 class Locked(object):
 
     def __init__(self, func, inst=None, class_=None, preconditions=()):
-        self.im_func = func
-        self.im_self = inst
-        self.im_class = class_
+        self.__func__ = func
+        self.__self__ = inst
+        self.__self_class__ = class_
         self.preconditions = preconditions
 
     def __get__(self, inst, class_):
-        return self.__class__(self.im_func, inst, class_, self.preconditions)
+        return self.__class__(
+            self.__func__, inst, class_, self.preconditions)
 
     def __call__(self, *args, **kw):
-        inst = self.im_self
+        inst = self.__self__
         if inst is None:
             inst = args[0]
-        func = self.im_func.__get__(self.im_self, self.im_class)
+        func = self.__func__.__get__(self.__self__, self.__self_class__)
 
         inst._lock_acquire()
         try:
